(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2017 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)


BeginPackage["AtomicDensityMatrix`Hamiltonian`", {
(*"PhysicalConstants`","Units`",*)
    "AtomicDensityMatrix`",
    "AtomicDensityMatrix`Common`",
    "AtomicDensityMatrix`CouplingCoefficients`",
    "AtomicDensityMatrix`TensorAlgebra`",
    "AtomicDensityMatrix`AtomicSystem`",
    "AtomicDensityMatrix`ExperimentalSystem`",
    "AtomicDensityMatrix`DensityMatrix`",
    "AtomicDensityMatrix`OpticalField`",
    "AtomicDensityMatrix`RotatingWaveApproximation`",
    "AtomicDensityMatrix`WignerEckart`",
    "AtomicDensityMatrix`WignerRotate`"
}];

Begin["`Private`"];

InternalHamiltonian[system : {_AtomicState..}] :=
    WignerEckart[system, {Energy, 0, 0}];

Energies[system : {_AtomicState..}, crit_ : True] :=
    Union[WignerEckart[#, {Energy, 0, 0}, #]& /@ SelectStates[system, crit]];

TransitionFrequencies[system : {_AtomicState..}, label1_ : True, label2_ : True] :=
    Transpose@Expand@Outer[Plus, Energies[system, label2], -Energies[system, label1]];

E1Hamiltonian[system : {_AtomicState..} /; AtomicSystemType[system] === "Toy",
    efield_ /; CartesianQ[efield] || ScalarQ[efield]] :=
    Module[{components = DeleteCases[EnsureList[efield], _?PossibleZeroQ]},
        If[
            components === {},
            WignerEckart[system, {Zero, 0, 0}],
            -ComplexExpand[Re[First@components]]WignerEckart[system, {Dipole, 1, 0}]
        ]
    ];

E1Hamiltonian[system : {_AtomicState..}, efield_?CartesianQ] :=
    -ComplexExpand[Re[efield]]\[CenterDot]ToCartesian[WignerEckart[system, {Dipole, 1}]];

MagneticHamiltonian[system : {_AtomicState..}, bf_?CartesianQ] :=
    -bf\[CenterDot]ToCartesian[WignerEckart[system, {MagneticMoment, 1}]];

MagneticHamiltonian[system : {_AtomicState..}, bf_?ScalarQ] := MagneticHamiltonian[system, {0, 0, bf}];

(*We define the polarizability operator of rank \[Kappa], denoted Subscript[\[Alpha], \[Kappa]], by*)
(*	Subscript[*)(*Subscript[H, Stark]={*)(*StyleBox["E",*)(*FontWeight->"Bold",*)(*FontSlant->"Plain"]\[CircleTimes]*)(*SuperscriptBox[*)(*StyleBox["E",*)(*FontWeight->"Bold",*)(*FontSlant->"Plain"],"*"]}, \[Kappa]]\[CenterDot]Subscript[\[Alpha], \[Kappa]] ,*)
(*where Subscript[H, Stark] is effective Stark Hamiltonian and E is the (complex) electric field. In terms of the polarizabilities, the diagonal elements for a z-directed field are given by*)
(*	Subscript[\[CapitalDelta], Stark]=-(E^2/2)(Subscript[\[Alpha], 0]+Subscript[\[Alpha], 2] (3m^2-j(j+1))/j(2j-1)).*)
(*Comparison of these equations allows the determination of the reduced matrix elements of Subscript[\[Alpha], \[Kappa]].*)

PolarizabilityHamiltonian[system : {_AtomicState..}, efield_?CartesianQ, k_] :=
    SphericalTensorProduct[ToCovariant@efield, ToCovariant@EC@efield, {k}]\[CenterDot]ECT@WignerEckart[system, {Polarizability, k}];

PolarizabilityHamiltonian[system : {_AtomicState..}, efield_?ScalarQ, k_] :=
    PolarizabilityHamiltonian[system, {0, 0, efield}, k];

PolarizabilityHamiltonian[system : {_AtomicState..}, efield_] :=
    Sum[PolarizabilityHamiltonian[system, efield, k], {k, 0, 2}];

(*TransitionPolarizabilityHamiltonian[system : {_AtomicState..}, opticalfield_?CartesianQ, staticfield_?CartesianQ, k_] :=*)
    (*SphericalTensorProduct[*)
        (*ToCovariant@staticfield,*)
        (*ToCovariant@ComplexExpand@Re@opticalfield, {k}*)
    (*]\[CenterDot]ECT@WignerEckart[system, {TransitionPolarizability, k}];*)

(*TransitionPolarizabilityHamiltonian[system : {_AtomicState..}, opticalfield_, staticfield_] :=*)
    (*Sum[TransitionPolarizabilityHamiltonian[system, opticalfield, staticfield, k], {k, 0, 2}];*)

MatrixElementMask[sys : {_AtomicState ..}, All] = 1;

MatrixElementMask[sys : {_AtomicState ..}, crit1_ -> crit2_] :=
    Module[{pos},
        pos = Alternatives @@@ (Flatten@StatePosition[sys, #] & /@ {crit1, crit2});
        SparseArray[{pos -> 1, Reverse@pos -> 1}, Length[sys] {1, 1}]
    ];

MatrixElementMask[sys : {_AtomicState ..}, trans_List] :=
    Clip[Total[MatrixElementMask[sys, #] & /@ trans], {0, 1}];

Options[Hamiltonian] = {ElectricField -> 0, MagneticField -> 0, Interaction -> Automatic, RestrictCoupling -> All};

Hamiltonian[system : {_AtomicState..}, expSystem : {_ExperimentalRegion..}, opts : OptionsPattern[]] :=
    Hamiltonian[system, #, opts]& /@ expSystem;

Hamiltonian[system : {_AtomicState..}, region_ExperimentalRegion, opts : OptionsPattern[]] :=
    Hamiltonian[system,
        ElectricField -> If[DefinedQ[region, ElectricField], ElectricField[region], {0, 0, 0}]
            + If[DefinedQ[region, OpticalParameters], OpticalField @@ (OpticalParameters[region]), {0, 0, 0}],
        MagneticField -> If[DefinedQ[region, MagneticField], MagneticField[region], {0, 0, 0}],
        opts
    ];

massageInteractionOption[type_] :=
    Switch[
        type,
        All,
        {"Internal", "ElectricDipole", "MagneticDipole", "Polarizability"},
        Automatic,
        {"Internal", "ElectricDipole", "MagneticDipole"},
        allButInternal,
        {"ElectricDipole", "MagneticDipole", "Polarizability"},
        automaticButInternal,
        {"ElectricDipole", "MagneticDipole"},
        None,
        {},
        _,
        EnsureList@type
    ];

Hamiltonian[system : {_AtomicState..}, fields : {{genericOptionsPattern[]}..}, opts : OptionsPattern[]] :=
    Total[
        Hamiltonian[system,
            Sequence @@ ({#, opts} /. {
                (Interaction -> All) -> (Interaction -> allButInternal),
                (Interaction -> Automatic) -> (Interaction -> automaticButInternal)
            })
        ]& /@ fields
    ] + Hamiltonian[system, opts];

Hamiltonian[system : {_AtomicState..}, OptionsPattern[]] :=
    Module[{ef = OptionValue[ElectricField], bf = OptionValue[MagneticField],
        type = massageInteractionOption@OptionValue@Interaction, zero = WignerEckart[system, {Zero, 0, 0}],
        int, e1, m1, pol, mask},
        If[ef === 0, ef = {0, 0, 0}];
        ef = If[MatchQ[ef, {{_?CartesianQ, _}..}], ef, {ef}];
        int = If[MemberQ[type, "Internal" | Internal],
            InternalHamiltonian[system],
            zero
        ];
        e1 = If[MemberQ[type, "ElectricDipole" | E1] && N@ef =!= {0., 0., 0.} && N@ef =!= 0.,
                Total[(E1Hamiltonian[system, #1]&) /@ ef],
                zero
        ];
        m1 = If[MemberQ[type, "MagneticDipole" | Magnetic] && N@bf =!= {0., 0., 0.} && N@bf =!= 0. && AtomicSystemType[system] =!= "Toy",
                MagneticHamiltonian[system, bf],
                zero
        ];
        pol = If[MemberQ[type, "Polarizability" | Polarizability],
                Total[(PolarizabilityHamiltonian[system, #1]&) /@ ef],
                zero
        ];
        mask = MatrixElementMask[system, OptionValue@RestrictCoupling];
        mask (int + e1 + m1 + pol)
    (*
        If[ MemberQ[type,"TransitionPolarizability"|TransitionPolarizability],
            TransitionPolarizabilityHamiltonian[system,Cases[ef,{x_,AC}->x,1][[1]],Cases[ef,{x_,DC}->x,1][[1]]],
            zero
        ]*)
    ];

End[];

EndPackage[];
