(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2017 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)


BeginPackage["AtomicDensityMatrix`Relaxation`", {
(*"PhysicalConstants`","Units`",*)
    "AtomicDensityMatrix`",
    "AtomicDensityMatrix`Common`",
    "AtomicDensityMatrix`CouplingCoefficients`",
    "AtomicDensityMatrix`TensorAlgebra`",
    "AtomicDensityMatrix`AtomicSystem`",
    "AtomicDensityMatrix`ExperimentalSystem`",
    "AtomicDensityMatrix`DensityMatrix`",
    "AtomicDensityMatrix`WignerEckart`",
    "AtomicDensityMatrix`WignerRotate`"}];

Begin["`Private`"];

IntrinsicRelaxation[system : {_AtomicState..}] = WignerEckart[system, {NaturalWidth, 0, 0}];

TransitRelaxation[system : {_AtomicState..}, gt_] := gt WignerEckart[system, {Identity, 0, 0}];

TransitRepopulation[system : {_AtomicState..}, gt_] := Module[
    {g = SublevelMultiplicity[GroundStates[system]]},
    gt WignerEckart[system, {GroundState, 0, 0}] / If[g == 0, 1, g]
];

Options[OpticalRepopulation] = Join[{HyperfineCoherenceTransfer -> False, ExpandDipoleRME -> True}, Options[DMVariables]];

OpticalRepopulation[system : {_AtomicState..}, opts : OptionsPattern[]] :=
    Module[{newopts},
        newopts = removeAutomaticOptions[OpticalRepopulation, opts];
        Switch[OptionValue[DensityMatrix, FilterRules[newopts, Options@DensityMatrix], Representation],
            "Zeeman", OpticalRepopulationZ[system, newopts],
            "PolarizationMoments" | PolarizationMoments, OpticalRepopulationPM[system, newopts]
        ]
    ];

OpticalRepopulationZ[system : {_AtomicState..}, opts : OptionsPattern[OpticalRepopulation]] :=
    Module[{d, branchingMat, degeneracyBoole, dm, rank},
        (*spherical components of the electric dipole operator*)
        d = SparseArray[Components@WignerEckart[system, {Dipole, 1}] /. ReducedME[_, {Dipole, 1}, _] -> 1];
        (* Matrix of branching rates *)
        branchingMat = SparseArray@Sqrt[(2 J[system] + 1) NaturalWidth[system] DirectSelfProduct[BranchingRatio[#2][#1] &, system]];
        (* Only allow transfer of coherences from a pair of degenerate levels to another pair of degenerate levels *)
        degeneracyBoole = SparseArray@DirectSelfProduct[
            Composition[Boole, PossibleZeroQ, Subtract],
            Energy@system + If[OptionValue[HyperfineCoherenceTransfer], 0, HyperfineShift@system]
        ];
        dm = DensityMatrix[system, FilterRules[{opts}, Options[DensityMatrix]]] degeneracyBoole;
        rank = Rank[d];
        Normal[degeneracyBoole Sum[(-1)^q (Transpose[branchingMat] d[[q + rank + 1]]).dm.(branchingMat d[[-q + rank + 1]]), {q, -rank, rank}]]
    ];

OpticalRepopulationPM[sys : {_AtomicState..}, opts : OptionsPattern[OpticalRepopulation]] /; AtomicSystemType[sys] == "Toy" :=
    Map[{{{#}}}&, OpticalRepopulationZ[sys, Representation -> "Zeeman", opts], {2}];

OpticalRepopulationPM[sys : {_AtomicState..}, opts : OptionsPattern[OpticalRepopulation]] :=
    Module[
        {levels, upperlevels, pm, le1, le2},
        levels = RemoveZeemanSublevels[sys];
        upperlevels = ExcitedStates@levels;
        pm = PolarizationMoments[upperlevels];
        If[OptionValue[ExpandDipoleRME], ExpandDipoleRME[sys, #], #]&@Table[
            4 / 3 Sum[
                le1 = upperlevels[[i1]];
                le2 = upperlevels[[i2]];
                If[
                    Abs[F[le1] - F[le2]] <= k <= F[le1] + F[le2],
                    (Energy[le1] - Energy[l1])^3 ReducedME[l1, {Dipole, 1}, le1] *
                        Component[pm[[i1, i2]], k] ReducedME[le2, {Dipole, 1}, l2] *
                            (-1)^(F[le1] + F[le2] + k + 1) SJS[{k, F[l2], F[l1]}, {1, F[le1], F[le2]}],
                    Transpose[{ConstantArray[0, 2k + 1]}]
                ],
                {i1, Length@upperlevels}, {i2, Length@upperlevels}
            ],
            {l1, levels}, {l2, levels}, {k, Abs[F[l1] - F[l2]], F[l1] + F[l2]}]
    ];

Options[OpticalRepopulationOperator] = {HyperfineCoherenceTransfer -> False, ExpandDipoleRME -> True};

OpticalRepopulationOperator[system : {_AtomicState..}, opts___Rule] :=
    Module[{dm, or},
        dm = DensityMatrix[system, ComplexExpandVariables -> False];
        or = OpticalRepopulation[system, ComplexExpandVariables -> False, opts];
        Normal@Transpose@Map[Partition[#, Length[system]]&, Last@CoefficientArrays[or, Flatten@dm], {2}]
    ];

(*The OpticalRepopulation matrix is equal to Map[Tr[#\[CenterDot]DensityMatrix[System]]&,OpticalRepopulationOperator[System],{2}]*)

Relaxation[as : {_AtomicState..}, es : {_ExperimentalRegion..}] :=
    (
        IntrinsicRelaxation[as]
            + TransitRelaxation[as, TransitRate[VoidRegion][#]]
            + TransitRelaxation[as, Sum[TransitRate[l][#], {l, es}]]
    )& /@ es;

Options@Repopulation = Join[
    {HyperfineCoherenceTransfer -> False, ExpandDipoleRME -> True, Method -> Automatic},
    FilterRules[Options@DMVariables, Except@DMLabel]
];

Repopulation[as : {_AtomicState..}, es : {_ExperimentalRegion..}, opts : OptionsPattern[]] :=
    (TransitRepopulation[as, Density[#] TransitRate[VoidRegion][#]] +
        OpticalRepopulation[as, DMLabel -> RegionLabel[#], FilterRules[{opts}, Options@OpticalRepopulation]] +
            Sum[
                Density[#] / Density[l] TransitRate[l][#] *
                    DensityMatrix[as, DMLabel -> RegionLabel[l], FilterRules[{opts}, Options@DensityMatrix]],
                {l, es}
            ]
    )& /@ es;

Options[FluorescenceOperator] = {ExpandDipoleRME -> True};

FluorescenceOperator[system : {_AtomicState..}, freq_, th_, phi_, opts___Rule] :=
    Module[{dc, gr, ikk},
        If[AtomicSystemType[system] == "Toy", Return[FluorescenceOperator[system, freq, opts]/(4 Pi)]];
        dc = ToCartesian@WignerEckart[system, {Dipole, 1}];
        dc = If[ExpandDipoleRME /. {opts} /. Options[OpticalRepopulation], ExpandDipoleRME[system, dc], dc];
        gr = DiagonalMatrix[Boole /@ GroundStateQ /@ system];
        ikk = IdentityMatrix[3] - DirectSelfProduct[{Cos[phi] Sin[th], Sin[phi] Sin[th], Cos[th]}];
        freq^4 / (2\[Pi]) Sum[dc[[i]].gr.dc[[j]] ikk[[i, j]], {i, 3}, {j, 3}]
    ];

FluorescenceOperator[system : {_AtomicState..}, freq_, opts___Rule] :=
    Module[{dc, gr},
        dc = ToCartesian@WignerEckart[system, {Dipole, 1}];
        dc = If[ExpandDipoleRME /. {opts} /. Options[OpticalRepopulation], ExpandDipoleRME[system, dc], dc];
        gr = DiagonalMatrix[Boole /@ GroundStateQ /@ system];
        4 / 3 freq^4 Sum[dc[[i]].gr.dc[[i]], {i, Length@dc}]
    ];

Options[Fluorescence] = {ExpandDipoleRME -> True};

dReduced[system_] := WignerEckart[system, {Dipole, 1}] /. ReducedME[_, {Dipole, 1}, _] -> 1;

branchingRatioMatrix[system_] := SparseArray@Sqrt[
    (2 J[system] + 1) NaturalWidth[system] DirectSelfProduct[BranchingRatio[#2][#1] &, system]
];

energyDifferenceMatrix[system_] := SparseArray@Sqrt[DirectSelfProduct[Subtract, Energy@system]];

degeneracyMatrix[system_] :=
    SparseArray@DirectSelfProduct[
        Composition[Boole, PossibleZeroQ, Subtract],
        Energy@system + HyperfineShift@system
    ];

(*Fluorescence[system : {_AtomicState..}, freq_, th_, phi_, opts___Rule] :=*)
    (*Tr[DensityMatrix[system].FluorescenceOperator[system, freq, th, phi, opts]];*)

Fluorescence[system : {_AtomicState..}, th_, phi_, opts___Rule] /; AtomicSystemType[system] == "Toy" :=
    1/(4 Pi) Fluorescence[system, opts];

Fluorescence[system : {_AtomicState..}, th_, phi_, opts___Rule] :=
    Module[{d, bMat, dm, dmopts, ikk},
        dmopts = FilterRules[{opts}, Options[DensityMatrix]];
        d = SparseArray@ToCartesian@dReduced[system];
        bMat = energyDifferenceMatrix[system] branchingRatioMatrix[system];
        dm = DensityMatrix[system, Sequence @@ dmopts] degeneracyMatrix[system];
        ikk = IdentityMatrix[3] - DirectSelfProduct[{Cos[phi] Sin[th], Sin[phi] Sin[th], Cos[th]}];
        3/(8 Pi) Sum[Tr[(Transpose@bMat d[[i]]).dm.(bMat d[[j]])] ikk[[i, j]], {i, 3}, {j, 3}]
    ];

Fluorescence[system : {_AtomicState..}, opts___Rule] :=
    Module[{d, bMat, dm, dmopts, rank},
        dmopts = FilterRules[{opts}, Options[DensityMatrix]];
        d = SparseArray@Components@dReduced[system];
        rank = Rank[d];
        bMat = energyDifferenceMatrix[system] branchingRatioMatrix[system];
        dm = DensityMatrix[system, Sequence @@ dmopts] degeneracyMatrix[system];
        Sum[(-1)^q Tr[(Transpose@bMat d[[q + rank + 1]]).dm.(bMat d[[-q + rank + 1]])], {q, -rank, rank}]
    ];

End[];

EndPackage[];
