(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2017 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)


Quiet[
    BeginPackage["AtomicDensityMatrix`WignerEckart`", {
        "Units`", (* Units are loaded because BohrMagneton is used below. Think about updating to allow version 9 units functionality. *)
        "AtomicDensityMatrix`",
        "AtomicDensityMatrix`Common`",
        "AtomicDensityMatrix`CouplingCoefficients`",
        "AtomicDensityMatrix`AtomicSystem`"
    }],
    General::obspkg
];


wignerEckartFactor::usage = "";

WignerEckart::usage = "";

ReducedME::usage = "";

OperatorOptions::usage = "";

EvaluateHyperfineShift::usage = "";

HyperfineShift::usage = "";

Identity::usage = "";

Zero::usage = "";

GroundState::usage = "";

ExcitedState::usage = "";

Polarization::usage = "";

StatisticalWeight::usage = "";

Dipole::usage = "";

ExpandDipoleRME::usage = "";

SpontaneousEmission::usage = "";

LandeGJ::usage = "";

LandeGF::usage = "";

DipoleExpansion::usage = "";

MagneticMoment::usage = "";

MagneticDipole::usage = "";

Begin["`Private`"];

(*Returns matrix elements between Zeeman sublevels in terms of reduced matrix elements. Just returns ReducedME for a toy system.*)

(*Wigner-Eckart theorem [Var88, Sec. 13.1.1, Eq. (2)]:*)
(**)
(* \[LeftAngleBracket]\[Gamma] F M|\!*)
(*\*SubsuperscriptBox[\(T\), \(q\), \(\[Kappa]\)]|\[Gamma]' F' M'\[RightAngleBracket]=(-1)^(F-M) ({*)
(* {F, \[Kappa], F'},*)
(* {-M, q, M'}*)
(*})(\[Gamma] F||T^\[Kappa]||\[Gamma]' F')*)
(* or*)
(*  \[LeftAngleBracket]\[Gamma] F M|\!*)
(*\*SubsuperscriptBox[\(T\), \(q\), \(\[Kappa]\)]|\[Gamma]' F' M'\[RightAngleBracket]=(-1)^(2q) (F'M'\[Kappa] q|F M) (\[Gamma] F||T^\[Kappa]||\[Gamma]' F')/Sqrt[2F+1]*)

(*General case*)

wignerEckartFactor[fm1 : {{_, _}..}, {k_, q_}, fm2 : {{_, _}..}] := wignerEckartFactor[fm1, {k, q}, fm2] =
    DirectProduct[wignerEckartFactor[#2, {k, q}, #1]&, fm1, fm2];

wignerEckartFactor[{f1_, m1_}, {k_, q_}, {f2_, m2_}] := wignerEckartFactor[{f1, m1}, {k, q}, {f2, m2}] =
    (2 f2 + 1)^(-1 / 2) CG[{f1, m1}, {k, q}, {f2, m2}];

(*The single component returned for a Toy system is wrapped in curly brackets *)
(*so that code will work properly when Toy systems are changed to real systems.*)
(*For example, a square matrix inside one pair of curly brackets is a zero-dimensional *)
(*Cartesian tensor, so ToCartesian returns the input. If we left off the brackets, then ToCartesian would choke on it.*)

WignerEckart[sys_, op_] := WignerEckart[sys, op, sys];

WignerEckart[s1 : ({_AtomicState..} | _AtomicState), {T_, k_Integer}, s2 : ({_AtomicState..} | _AtomicState)] /; AtomicSystemType[s1] === "Toy" := {ReducedME[s1, {T, k}, s2]};

WignerEckart[sys1 : {_AtomicState..}, {T_, k_Integer, q_}, sys2 : {_AtomicState..}] /; AtomicSystemType[sys1] === "Toy" := ReducedME[sys1, {T, k}, sys2];

WignerEckart[s1_AtomicState, {T_, k_Integer}, s2_AtomicState] := WignerEckart[{s1}, {T, k}, {s2}][[All, All, 1, 1]];

WignerEckart[s1_AtomicState, {T_, k_Integer, q_}, s2_AtomicState] := WignerEckart[{s1}, {T, k, q}, {s2}][[1, 1]];

WignerEckart[sys1 : {_AtomicState..}, {T_, k_Integer}, sys2 : {_AtomicState..}] := WignerEckart[sys1, {T, k, Range[k, -k, -1]}, sys2];

WignerEckart[sys1 : {_AtomicState..}, {T_, k_Integer, q_}, sys2 : {_AtomicState..}] := WignerEckart[sys1, {T, k, {q}}, sys2][[1, 1]];

WignerEckart[sys1 : {_AtomicState..}, {T_, k_Integer, qlist_List}, sys2 : {_AtomicState..}] := Module[{fm1, fm2, rme},
    {fm1, fm2} = Transpose /@ Through /@ {F, M} /@ {sys1, sys2};
    rme = ReducedME[sys1, {T, k}, sys2];
    Expand[{rme wignerEckartFactor[fm1, {k, #}, fm2] & /@ qlist}]
];

(*Faster version for known diagonal operators. Maybe should make a global variable listing diagonal operators and check against that instead. *)
(*WignerEckart[sys : {_AtomicState..}, {T : Energy | Zero, 0, 0}] :=*)
    (*DiagonalMatrix[WignerEckart[#, {T, 0, 0}, #]& /@ sys];*)

WignerEckart[label_, j_?HalfIntegerQ, Tn_] := WignerEckart[ZeemanSystem[label, j], Tn];

WignerEckart[j_?HalfIntegerQ, Tn_] := WignerEckart[ZeemanSystem[j], Tn];

(*Reduced matrix elements*)

ReducedME[sys : {_AtomicState..}, {op_, k_}] := ReducedME[sys, {op, k}, sys];

ReducedME[sys1 : {_AtomicState..}, {op_, k_}, sys2 : {_AtomicState..}] := ArrayFlatten[
    DirectProduct[ConstantArray[1, {##}]&, CountZeemanSublevels@sys1, CountZeemanSublevels@sys2] *
        DirectProduct[ReducedME[#1, {op, k}, #2] &, RemoveZeemanSublevels@sys1, RemoveZeemanSublevels@sys2]
];

(*Polarization operator*)

ReducedME[s1_AtomicState, {Polarization, k_Integer}, s2_AtomicState] := Boole[StateLabel[s1] === StateLabel[s2]] Sqrt[2k + 1];

ReducedME[s1_AtomicState, {Polarization[{s1p_, f1_?NumericQ}, {s2p_, f2_?NumericQ}], k_Integer}, s2_AtomicState] :=
    Boole[{StateLabel[s1], StateLabel[s2], F[s1], F[s2]} === {s1p, s2p, f1, f2}] Sqrt[2k + 1];

ReducedME[s1_AtomicState, {Polarization[s1p_, s2p_], k_Integer}, s2_AtomicState] :=
    Boole[{StateLabel[s1], StateLabel[s2]} === {s1p, s2p}] Sqrt[2k + 1];

(*Hyperfine*)

(*Special case, statistical weight operator*)

ReducedME[s1_AtomicState, {StatisticalWeight, 0}, s2_AtomicState] /; AtomicSystemType[s1] === "HyperfineZeeman" :=
    With[{J1 = J@s1, I1 = NuclearSpin@s1, F1 = F@s1, J2 = J@s2, I2 = NuclearSpin@s2, F2 = F@s2},
        (-1)^(J1 + I1 + F2) Sqrt[(2 F1 + 1)(2 F2 + 1)] / (2 I1 + 1) SJS[{J1, F1, I1}, {F2, J2, 0}] *
            ReducedME[DeleteParameter[s1, F, NuclearSpin], {StatisticalWeight, 0}, DeleteParameter[s2, F, NuclearSpin]]
    ];

(*The energy of a given hyperfine level*)

hyperfineK[j_, i_, f_] = f(f + 1) - i(i + 1) - j(j + 1);

EvaluateHyperfineShift[expr_] := expr /. HyperfineShift[j_, i_, f_, a_, b_] :> With[
    {k = hyperfineK[j, i, f]},
    1 / 2 a k + Piecewise[{{
        b (3 / 2 k (k + 1) - 2 i (i + 1) j (j + 1)) / (2 i (2 i - 1) 2 j (2 j - 1)),
        j > 1 / 2 && i > 1 / 2
    }}]
];

HyperfineShift[s_AtomicState] /; AtomicSystemType[s] === "HyperfineZeeman" :=
    EvaluateHyperfineShift@HyperfineShift[J[s], NuclearSpin[s], F[s], HyperfineA[s], HyperfineB[s]];

HyperfineShift[s_AtomicState] := 0;

HyperfineShift[s : {_AtomicState..}] := HyperfineShift /@ s;

hyperfineRMEFactor[s1_, k_Integer, s2_] := hyperfineRMEFactor[s1, k, s2] =
    (-1)^(J[s1] + NuclearSpin[s1] + F[s2] + k) Sqrt[(2F[s1] + 1)(2F[s2] + 1)] *
        SJS[{J[s1], F[s1], NuclearSpin[s1]}, {F[s2], J[s2], k}];

ReducedME[s1_AtomicState, {Energy, 0}, s2_AtomicState] /; AtomicSystemType[s1] === "HyperfineZeeman" :=
    With[{sJ = DeleteParameter[{s1, s2}, F, NuclearSpin]},
        hyperfineRMEFactor[s1, 0, s2] *
            (ReducedME[sJ[[1]], {Energy, 0}, sJ[[2]]] + HyperfineShift[s1] ReducedME[sJ[[1]], {Identity, 0}, sJ[[2]]])
    ];

(* ReducedME for hyperfine structure, assuming the operator commutes with nuclear spin *)
(* (I J_1 F_1||T^k||I J_2 F_2) = (-1)^(J_1+I+F_2+k) Sqrt[(2 F_1+1)(2 F_2+1)] SJS[{J_1,F_1,I},{F_2,J_2,k}] (J_1||T^k||J_2) *)
(* Sobelman Eq. (4.175), p. 84 *)

ReducedME[s1_AtomicState, {T_, k_Integer}, s2_AtomicState] /; AtomicSystemType[s1] === "HyperfineZeeman" :=
    ReducedME[s1, {T, k}, s2] =
        hyperfineRMEFactor[s1, k, s2] *
            ReducedME[DeleteParameter[s1, F, NuclearSpin], {T, k}, DeleteParameter[s2, F, NuclearSpin]];

(*Zero*)

ReducedME[s1_AtomicState, {Zero, 0}, s2_AtomicState] = 0;

(*Identity operator*)

(*[Var88, Sec. 13.1.1, Eq. (3)]*)

ReducedME[s1_AtomicState, {Identity, 0}, s2_AtomicState] :=
    Boole[StateLabel[s1] === StateLabel[s2]] Sqrt[2J[s1] + 1];

(*Ground state operator*)

ReducedME[s1_AtomicState, {GroundState, 0}, s2_AtomicState] :=
    Boole@GroundStateQ[s1] ReducedME[s1, {Identity, 0}, s2];

(*Excited state operator*)

ReducedME[s1_AtomicState, {ExcitedState, 0}, s2_AtomicState] :=
    Boole@Not@GroundStateQ[s1] ReducedME[s1, {Identity, 0}, s2];

(*Natural width operator*)

ReducedME[s1_AtomicState, {NaturalWidth, 0}, s2_AtomicState] :=
    NaturalWidth[s1] ReducedME[s1, {Identity, 0}, s2];

(*Statistical weight operator*)

ReducedME[s1_AtomicState, {StatisticalWeight, 0}, s2_AtomicState] :=
    1 / (2J[s1] + 1) ReducedME[s1, {Identity, 0}, s2];

(*Unperturbed energy operator*)

ReducedME[s1_AtomicState, {Energy, 0}, s2_AtomicState] :=
    Energy[s1]ReducedME[s1, {Identity, 0}, s2];

(*E1 dipole operator*)

(*(\[Gamma] J||d||\[Gamma]'J')=0 for same-parity levels*)

ReducedME[s1_AtomicState, {Dipole, 1}, s2_AtomicState] /;
    Parity[s1] === Parity[s2] || StateLabel[s1] === StateLabel[s2] :=
    ReducedME[s1, {Dipole, 1}, s2] = 0;

(*Reorder states, using (\[Gamma] J||d||\[Gamma]'J')=(-1)^(J-J') (\[Gamma]'J'||d||\[Gamma] J) *)
(*If states are in the right order, replace expression with a placeholder whose value can be set later*)

ReducedME[s1_AtomicState, {Dipole, 1}, s2_AtomicState] :=
    ReducedME[s1, {Dipole, 1}, s2] =
        If[OrderedQ[{StateLabel[s1], StateLabel[s2]}],
            ReducedME[StateLabel[s1], {Dipole, 1}, StateLabel[s2]],
            If[AtomicSystemType[s1] === "Toy", 1, (-1)^(J[s1] - J[s2])] *
                ReducedME[StateLabel[s2], {Dipole, 1}, StateLabel[s1]]
        ];

DipoleExpansion[s1_AtomicState, s2_AtomicState] := Module[
    {S1, L1, J1, S2, L2, J2, Lab1, Lab2, W1, W2, G1, G2, BR12, BR21},
    {{S1, L1, J1, W1, Lab1, G1}, {S2, L2, J2, W2, Lab2, G2}} = Through /@ {S, L, J, Energy, StateLabel, NaturalWidth} /@ {s1, s2};
    {BR12, BR21} = {BranchingRatio[s1][s2], BranchingRatio[s2][s1]};
    ReducedME[Lab1, {Dipole, 1}, Lab2] ->
        Which[
            BR21 === 0 && BR12 =!= 0,
            Sqrt[3] / 2 Sqrt[(2 J2 + 1) G2 BR12 / (W2 - W1)^3],
            BR12 === 0 && BR21 =!= 0,
            Sqrt[3] / 2 Sqrt[(2 J1 + 1) G1 BR21 / (W1 - W2)^3],
            True,
            ReducedME[Lab1, {Dipole, 1}, Lab2]
        ]
];

ExpandDipoleRME[syst_, expr_] :=
    expr /. Flatten@DirectSelfProduct[DipoleExpansion, Union[syst, SameTest -> (StateLabel[#1] === StateLabel[#2]&)]];

(*Spontaneous emission*)
(*
ReducedME[s1_AtomicState, {SpontaneousEmission, 1}, s2_AtomicState] /; StateLabel[s1] === StateLabel[s2] = 0;

ReducedME[s1_AtomicState, {SpontaneousEmission, 1}, s2_AtomicState] /; Parity[s1] === Parity[s2] = 0;

ReducedME[s1_AtomicState, {SpontaneousEmission, 1}, s2_AtomicState] /; !OrderedQ[{StateLabel[s1], StateLabel[s2]}] :=
    If[AtomicSystemType[s1] === "Toy", 1, (-1)^(J[s1] - J[s2])] ReducedME[s2, {SpontaneousEmission, 1}, s1];

ReducedME[s1_AtomicState, {SpontaneousEmission, 1}, s2_AtomicState] :=
    If[LevelPosition[s1] === Lower,
        Sqrt[(2J[s2] + 1)NaturalWidth[s2]BranchingRatio[StateLabel[s1]][s2]],
        Sqrt[(2J[s1] + 1)NaturalWidth[s1]BranchingRatio[StateLabel[s2]][s1]]
    ];
*)

(*Angular momentum operator*)

ReducedME[s1_AtomicState, {J, 1}, s2_AtomicState] :=
    Sqrt[J[s1](J[s1] + 1)(2J[s1] + 1)] Boole[StateLabel[s1] === StateLabel[s2]];

ReducedME[s1_AtomicState, {S, 1}, s2_AtomicState] := Module[
    {S1, L1, J1, S2, L2, J2, Lab1, Lab2},
    {{S1, L1, J1, Lab1}, {S2, L2, J2, Lab2}} = Through /@ {S, L, J, StateLabel} /@ {s1, s2};
    Which[
    (*Coupling within the same J-state, evaluate explicitly*)
        Lab1 === Lab2,
        If[J1 < 1 / 2,
            0,
            Sqrt[J1 (J1 + 1) (2 J1 + 1)] (J1 (J1 + 1) + S1 (S1 + 1) - L1 (L1 + 1)) / (2 J1 (J1 + 1))
        ],
    (*Coupling between fine-structure states*)
        FineStructureStatesQ[s1, s2],
        If[DefinedQ[{s1, s2}, {S, L, J}],
        (*If all quantum numbers are defined, evaluate explicitly*)
            If[S1 =!= S2 || L1 =!= L2,
                0,
                (-1)^(S1 + L1 + J2 + 1) Sqrt[S1 (S1 + 1) (2 S1 + 1) (2 J1 + 1) (2 J2 + 1)] SJS[{S1, J1, L1}, {J2, S1, 1}]
            ],
        (*Otherwise return symbolic placeholder*)
            ReducedME[Lab1, {S, 1}, Lab2]
        ],
    (*If coupling is not between fine-structure states, return zero*)
        True,
        0
    ]
];

ReducedME[s1_AtomicState, {L, 1}, s2_AtomicState] := Module[
    {S1, L1, J1, S2, L2, J2, Lab1, Lab2},
    {{S1, L1, J1, Lab1}, {S2, L2, J2, Lab2}} = Through /@ {S, L, J, StateLabel} /@ {s1, s2};
    Which[
    (*Coupling within the same J-state, evaluate explicitly*)
        Lab1 === Lab2,
        If[J1 < 1 / 2,
            0,
            Sqrt[J1 (J1 + 1) (2 J1 + 1)] (J1 (J1 + 1) + L1 (L1 + 1) - S1 (S1 + 1)) / (2 J1 (J1 + 1))
        ],
    (*Coupling between fine-structure states*)
        FineStructureStatesQ[s1, s2],
        If[DefinedQ[{s1, s2}, {S, L, J}],
        (*If all quantum numbers are defined, evaluate explicitly*)
            If[S1 =!= S2 || L1 =!= L2,
                0,
                (-1)^(S1 + L1 + J2 + 1) Sqrt[L1 (L1 + 1) (2 L1 + 1) (2 J1 + 1) (2 J2 + 1)] SJS[{L1, J1, S1}, {J2, L1, 1}]
            ],
        (*Otherwise return symbolic placeholder*)
            ReducedME[Lab1, {L, 1}, Lab2]
        ],
    (*If coupling is not between fine-structure states, return zero*)
        True,
        0
    ]
];

(*Magnetic moment operator*)

(*g factors*)

LandeGJ[0, l_, s_] = 0;
LandeGJ[j_, l_, s_] = 1 + (j (j + 1) + s (s + 1) - l (l + 1)) / (2 j (j + 1));
LandeGJ[s_AtomicState] := LandeGJ[J[s], L[s], S[s]];
SetAttributes[LandeGJ, Listable];

LandeGF[0, j_, l_, s_, i_] = 0;
LandeGF[f_, j_, l_, s_, i_] = LandeGJ[j, l, s] (f (f + 1) + j (j + 1) - i (i + 1)) / (2 f (f + 1));
LandeGF[s_AtomicState] := LandeGF[F[s], J[s], L[s], S[s], NuclearSpin[s]];
SetAttributes[LandeGF, Listable];

ReducedME[s1_AtomicState, {MagneticMoment, 1}, s2_AtomicState] :=
    -BohrMagneton Which[
    (*For coupling within the same state, calculate using g-factor for efficiency*)
        StateLabel[s1] === StateLabel[s2],
        Which[
            DefinedQ[s1, GFactor],
            GFactor[s1],
            DefinedQ[s1, {J, L, S}],
            LandeGJ[s1],
            True,
            GFactor[s1]
        ] ReducedME[s1, {J, 1}, s2],
    (*Between two fine-structure states, calculate using L and S*)
        FineStructureStatesQ[s1, s2],
        ReducedME[s1, {L, 1}, s2] + 2 ReducedME[s1, {S, 1}, s2],
    (*Otherwise, return zero*)
        True,
        0
    ];

(*MagneticDipole operator*)

(*(\[Gamma] J||d||\[Gamma]'J')=0 for same-parity levels*)

(*ReducedME[s1_AtomicState,{MagneticDipole,1},s2_AtomicState] /;
	Parity[s1]!=Parity[s2]||StateLabel[s1]==StateLabel[s2] := 0*)

(*Reorder states, using (s1 J||MagneticDipole||\[Gamma]'J')=(-1)^(J-J') (\[Gamma]'J'||MagneticDipole||\[Gamma] J) *)
(* NEED TO CHECK IF THIS IS CORRECT!*)

(*ReducedME[s1_AtomicState,{MagneticDipole,1},s2_AtomicState] /; 
	!OrderedQ[{StateLabel[s1],StateLabel[s2]}] :=
		(-1)^(J[s1]-J[s2]) ReducedME[s2,{MagneticDipole,1},s1]*)

(*Polarizability operator*)

(*We define the polarizability operator of rank \[Kappa], denoted Subscript[\[Alpha], \[Kappa]], by*)
(*	Subscript[\!\(TraditionalForm\`*)
(*\*SubscriptBox[\(H\), \(Stark\)] = {\**)
(*StyleBox["E",*)
(*FontWeight->"Bold",*)
(*FontSlant->"Plain"]\[CircleTimes]\**)
(*SuperscriptBox[*)
(*StyleBox["E",*)
(*FontWeight->"Bold",*)
(*FontSlant->"Plain"], "*"]}\), \[Kappa]]\[CenterDot]Subscript[\[Alpha], \[Kappa]] ,*)
(*where Subscript[H, Stark] is effective Stark Hamiltonian and E is the (complex) electric field. In terms of the polarizabilities, the diagonal elements for a z-directed field are given by*)
(*	Subscript[\[CapitalDelta], Stark]=-(E^2/2)(Subscript[\[Alpha], 0]+Subscript[\[Alpha], 2] (3m^2-j(j+1))/j(2j-1)).*)
(*Comparison of these equations allows the determination of the reduced matrix elements of Subscript[\[Alpha], \[Kappa]].*)

alphaFactor[k_Integer, j_?HalfIntegerQ] /; k >= 2j + 1 := 0;

alphaFactor[k_Integer, j_?HalfIntegerQ] := alphaFactor[k, j] =
    Simplify[
        -(1 / 2)(Sqrt[2 j + 1]) / (CG[{1, Mod[k, 2]}, {1, -Mod[k, 2]}, {k, 0}] CG[{j, j}, {k, 0}, {j, j}]),
        Element[2 j, Integers]
    ];

ReducedME[s1_AtomicState, {Polarizability, k_Integer}, s2_AtomicState] :=
    alphaFactor[k, J[s1]] Polarizability[k][s1] Boole[StateLabel[s1] === StateLabel[s2]];

(*TransitionPolarizability operator*)

(*ReducedME[s1_AtomicState, {TransitionPolarizability, k_Integer}, s2_AtomicState] :=*)
(*Switch[k, 0, -Sqrt[3], 1, Sqrt[2], 2, 1]TransitionPolarizability[k][s1, s2] **)
(*(1 - Boole[StateLabel[s1] === StateLabel[s2]]);*)

End[];

EndPackage[];
